"""A preprocessor that extracts all of the attachments from the notebook file.

The extracted attachments are returned in the 'resources' dictionary.

Based on the ExtractOutputsProcessor in nbconvert... the license for
nbconvert is

# Licensing terms

This project is licensed under the terms of the Modified BSD License
(also known as New or Revised or 3-Clause BSD), as follows:

- Copyright (c) 2001-2015, IPython Development Team
- Copyright (c) 2015-, Jupyter Development Team

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

Neither the name of the Jupyter Development Team nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

## About the Jupyter Development Team

The Jupyter Development Team is the set of all contributors to the Jupyter project.
This includes all of the Jupyter subprojects.

The core team that coordinates development on GitHub can be found here:
https://github.com/jupyter/.

## Our Copyright Policy

Jupyter uses a shared copyright model. Each contributor maintains copyright
over their contributions to Jupyter. But, it is important to note that these
contributions are typically only changes to the repositories. Thus, the Jupyter
source code, in its entirety is not the copyright of any single person or
institution. Instead, it is the collective copyright of the entire Jupyter
Development Team. If individual contributors want to maintain a record of what
changes/contributions they have specific copyright on, they should indicate
their copyright in the commit message of the change, when they commit the
change to one of the Jupyter repositories.

With this in mind, the following banner should be used in any source code file
to indicate the copyright and license terms:

    # Copyright (c) Jupyter Development Team.
    # Distributed under the terms of the Modified BSD License.
"""

from binascii import a2b_base64
import sys
import os
import re
from typing import Tuple, TYPE_CHECKING

from traitlets import Unicode, Set
from nbconvert.preprocessors.base import Preprocessor

if TYPE_CHECKING:
    from nbformat import NotebookNode  # noqa: F401 # only imported for type checking


class ExtractAttachmentsPreprocessor(Preprocessor):  # type: ignore # no types available
    """
    Extracts all of the outputs from the notebook file.

    The extracted outputs are returned in the 'resources' dictionary.
    """

    output_filename_template = Unicode("{unique_key}_{cell_index}_{name}").tag(
        config=True
    )

    extract_output_types = Set(
        {"image/png", "image/jpeg", "image/svg+xml", "application/pdf"}
    ).tag(config=True)

    def preprocess_cell(
        self, cell: "NotebookNode", resources: dict, cell_index: int
    ) -> Tuple["NotebookNode", dict]:
        """Apply a transformation on each cell.

        Parameters
        ----------
        cell : NotebookNode cell
            Notebook cell being processed
        resources : dictionary
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.
        cell_index : int
            Index of the cell being processed (see base.py)

        """
        # Get files directory if it has been specified
        output_files_dir = resources.get("output_files_dir", None)

        # Make sure outputs key exists
        if not isinstance(resources["outputs"], dict):
            resources["outputs"] = {}

        # Loop through all of the attachments in the cell
        for name, attach in cell.get("attachments", {}).items():
            orig_name = name
            name = re.sub(r"%[\w\d][\w\d]", "-", name)
            for mime, data in attach.items():
                if mime not in self.extract_output_types:
                    continue

                # Binary files are base64-encoded, SVG is already XML
                if mime in {"image/png", "image/jpeg", "application/pdf"}:
                    # data is b64-encoded as text (str, unicode),
                    # we want the original bytes
                    data = a2b_base64(data)
                elif sys.platform == "win32":
                    data = data.replace("\n", "\r\n").encode("UTF-8")
                else:
                    data = data.encode("UTF-8")

                filename = self.output_filename_template.format(
                    cell_index=cell_index,
                    name=name,
                    unique_key=resources.get("unique_key", ""),
                )

                if output_files_dir is not None:
                    filename = os.path.join(output_files_dir, filename)

                if name.endswith(".gif") and mime == "image/png":
                    filename = filename.replace(".gif", ".png")

                # In the resources, make the figure available via
                #   resources['outputs']['filename'] = data
                resources["outputs"][filename] = data

                # now we need to change the cell source so that it links to the
                # filename instead of `attachment:`
                attach_str = "attachment:" + orig_name
                if attach_str in cell.source:
                    cell.source = cell.source.replace(attach_str, filename)

        return cell, resources



"""Test the convert_thermo_hw module."""
import os
import pkg_resources

import nbformat
from thermohw.convert_thermo_hw import pdf_exp, nb_exp


def test_convert_pathological_image_name() -> None:
    """Test that converting a notebook with a pathological image name works."""
    filename = os.path.join("test-pathological-image-name.ipynb")
    filename = pkg_resources.resource_filename(__name__, filename)
    res = {
        "unique_key": "test-pathological-image-name",
        "by_hand": False,
        "remove_solution": False,
    }
    solution_pdf, _ = pdf_exp.from_filename(filename, res)
    assert len(solution_pdf) > 0


def test_tags() -> None:
    """Test tag support."""
    filename = os.path.join("test-cell-tags.ipynb")
    filename = pkg_resources.resource_filename(__name__, filename)
    res = {"remove_solution": True, "legacy": False}
    problem_nb = nbformat.read(filename, as_version=4)
    if "celltoolbar" in problem_nb.metadata:
        del problem_nb.metadata["celltoolbar"]
    solution_nb, _ = nb_exp.from_notebook_node(problem_nb, res)
    assert len(solution_nb) > 0



"""
Convert a Jupyter Notebook-format homework problem.

The single input Jupyter Notebook is converted into four files:

- A Notebook without the solution for students to fill in
- A PDF without the solution for students to fill in
- A Notebook with the solution
- A PDF with the solution

The solution is detected as any cells below a Markdown cell with the phrase "##
solution" in the cell. Below the solution-delimiting cell, any Markdown cell
that starts with a level-three header (###) is retained as a divider between
sections of the solution.

Input files should be named according to the convention::

    homework-A-B.ipynb

where ``A`` is the homework number and ``B`` is the problem number of this input
file in the homework assignment.

Methods
-------
process(hw_num, problems_to_do=None, prefix=None): Process the files for
    homework number ``hw_num``. Only process the specific problems in the
    ``problems`` argument.

main(argv=None): Process the command line arguments and run the `process`
    function

"""
# Standard library
from typing import Iterable, Dict, Sequence, Optional, List, Any, Union
from pathlib import Path
from argparse import ArgumentParser
from zipfile import ZipFile
from io import BytesIO
from datetime import date
import shutil
import sys

# Third Party
from nbconvert import NotebookExporter, PDFExporter
from traitlets.config import Config
from nbconvert.writers import FilesWriter
import nbformat

# Local imports
from .extract_attachments import ExtractAttachmentsPreprocessor
from .pymarkdown import PyMarkdownPreprocessor
from .preprocessors import RawRemover, SolutionRemover
from .filters import convert_div, convert_raw_html
from .utils import combine_pdf_as_bytes

c = Config()
here = Path(__file__).resolve().parent
c.PDFExporter.template_file = str(here / "homework.tpl")
c.PDFExporter.filters = {
    "convert_div": convert_div,
    "convert_raw_html": convert_raw_html,
}
c.PDFExporter.latex_count = 1


nb_exp = NotebookExporter(
    preprocessors=[RawRemover, SolutionRemover, PyMarkdownPreprocessor]
)

pdf_exp = PDFExporter(
    preprocessors=[
        RawRemover,
        SolutionRemover,
        PyMarkdownPreprocessor,
        ExtractAttachmentsPreprocessor(config=c),
    ],
    config=c,
)
pdf_exp.writer.build_directory = "."


def process(
    hw_num: int,
    problems_to_do: Optional[Iterable[int]] = None,
    prefix: Optional[Path] = None,
    by_hand: Optional[Iterable[int]] = None,
    legacy: bool = False,
) -> None:
    """Process the homework problems in ``prefix`` folder.

    Arguments
    ---------
    hw_num
        The number of this homework
    problems_to_do, optional
        A list of the problems to be processed
    prefix, optional
        A `~pathlib.Path` to this homework assignment folder
    by_hand, optional
        A list of the problems that should be labeled to be completed
        by hand and have an image with the solution included.
    legacy, optional
        A boolean flag determining whether the legacy method of finding
        solutions will be used, based on parsing cell content.
    """
    if prefix is None:
        prefix = Path(".")

    problems: Iterable[Path]

    if problems_to_do is None:
        # The glob syntax here means a the filename must start with
        # homework-, be followed the homework number, followed by a
        # dash, then a digit representing the problem number for this
        # homework number, then any number of characters (in practice
        # either nothing or, rarely, another digit), then the ipynb
        # extension. Examples:
        # homework-1-1.ipynb, homework-10-1.ipynb, homework-3-10.ipynb
        problems = list(prefix.glob(f"homework-{hw_num}-[0-9]*.ipynb"))
    else:
        problems = [prefix / f"homework-{hw_num}-{i}.ipynb" for i in problems_to_do]

    problems = sorted(problems, key=lambda k: k.stem[-1])

    output_directory: Path = (prefix / "output").resolve()
    fw = FilesWriter(build_directory=str(output_directory))

    assignment_zip_name = output_directory / f"homework-{hw_num}.zip"
    solution_zip_name = output_directory / f"homework-{hw_num}-soln.zip"

    assignment_pdfs: List[BytesIO] = []
    solution_pdfs: List[BytesIO] = []

    assignment_pdf: bytes
    solution_pdf: bytes
    assignment_nb: str
    solution_nb: str

    res: Dict[str, Union[Dict[str, bool], str, bool]] = {
        "delete_pymarkdown": True,
        "global_content_filter": {"include_raw": False},
        "legacy": legacy,
    }

    for problem in problems:
        print("Working on:", problem)
        res["unique_key"] = problem.stem
        problem_number = int(problem.stem.split("-")[-1])
        if by_hand is not None and problem_number in by_hand:
            res["by_hand"] = True
        else:
            res["by_hand"] = False
        problem_fname = str(problem.resolve())
        problem_nb = nbformat.read(problem_fname, as_version=4)
        if "celltoolbar" in problem_nb.metadata:
            del problem_nb.metadata["celltoolbar"]

        # Process assignments
        res["remove_solution"] = True
        assignment_pdf, _ = pdf_exp.from_notebook_node(problem_nb, resources=res)
        assignment_pdfs.append(BytesIO(assignment_pdf))

        assignment_nb, _ = nb_exp.from_notebook_node(problem_nb, resources=res)
        with ZipFile(assignment_zip_name, mode="a") as zip_file:
            zip_file.writestr(problem.name, assignment_nb)

        # Process solutions
        res["remove_solution"] = False
        solution_pdf, _ = pdf_exp.from_notebook_node(problem_nb, resources=res)
        solution_pdfs.append(BytesIO(solution_pdf))

        solution_nb, _ = nb_exp.from_notebook_node(problem_nb, resources=res)
        with ZipFile(solution_zip_name, mode="a") as zip_file:
            zip_file.writestr(problem.stem + "-soln" + problem.suffix, solution_nb)

    resources: Dict[str, Any] = {
        "metadata": {
            "name": f"homework-{hw_num}",
            "path": str(prefix),
            "modified_date": date.today().strftime("%B %d, %Y"),
        },
        "output_extension": ".pdf",
    }
    fw.write(combine_pdf_as_bytes(assignment_pdfs), resources, f"homework-{hw_num}")

    resources["metadata"]["name"] = f"homework-{hw_num}-soln"
    fw.write(combine_pdf_as_bytes(solution_pdfs), resources, f"homework-{hw_num}-soln")


def main(argv: Optional[Sequence[str]] = None) -> None:
    """Parse arguments and process the homework assignment."""
    parser = ArgumentParser(description="Convert Jupyter Notebook assignments to PDFs")
    parser.add_argument(
        "--hw",
        type=int,
        required=True,
        help="Homework number to convert",
        dest="hw_num",
    )
    parser.add_argument(
        "-p",
        "--problems",
        type=int,
        help="Problem numbers to convert",
        dest="problems",
        nargs="*",
    )
    parser.add_argument(
        "--by-hand",
        type=int,
        help="Problem numbers to be completed by hand",
        dest="by_hand",
        nargs="*",
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        help=(
            "Remove the output folder. If --problems is passed, the specified"
            "problems will be processed. Otherwise, will exit after cleaning."
        ),
    )
    parser.add_argument(
        "--legacy",
        action="store_true",
        help=(
            "Enable legacy behavior for finding solution cells, based on "
            "finding specific cell content."
        ),
    )
    args = parser.parse_args(argv)
    prefix = Path(f"homework/homework-{args.hw_num}")
    if args.clean:
        shutil.rmtree(prefix.joinpath("output"), ignore_errors=True)
        if not args.problems:
            sys.exit(0)

    process(
        args.hw_num,
        args.problems,
        prefix=prefix,
        by_hand=args.by_hand,
        legacy=args.legacy,
    )


if __name__ == "__main__":
    main()



"""Convert thermo homework assignments."""

from .convert_thermo_hw import process as hw_process  # noqa: F401
from .extract_attachments import ExtractAttachmentsPreprocessor  # noqa: F401
from .pymarkdown import PyMarkdownPreprocessor  # noqa: F401
from .preprocessors import RawRemover, SolutionRemover  # noqa: F401
from .filters import (  # noqa: F401
    ALLOWED_ALERT_TYPES,
    div_filter,
    convert_div,
    raw_html_filter,
    convert_raw_html,
)
from ._version import __version__  # noqa: F401



"""Nbconvert preprocessor for the python-markdown nbextension.

This module is copied from the ipython-contrib/jupyter_contrib_nbextensions
package. The original source of the module is available at

https://github.com/ipython-contrib/jupyter_contrib_nbextensions/blob/master/src/jupyter_contrib_nbextensions/nbconvert_support/pre_pymarkdown.py

The jupyter_contrib_nbextensions package is licensed as follows:

====================================
 The IPython-contrib licensing terms
====================================

IPython-contrib is licensed under the terms of the Modified BSD License (also
known as New or Revised or 3-Clause BSD), as follows:

- Copyright (c) 2013-2015, IPython-contrib Developers

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

Neither the name of the IPython-contrib Developers nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from typing import TYPE_CHECKING, Tuple, Dict
import re

from nbconvert.preprocessors import Preprocessor

if TYPE_CHECKING:
    from nbformat import NotebookNode  # noqa: F401 # typing only import


class PyMarkdownPreprocessor(Preprocessor):  # type: ignore
    """
    :mod:`nbconvert` Preprocessor for the python-markdown nbextension.

    This :class:`~nbconvert.preprocessors.Preprocessor` replaces kernel code in
    markdown cells with the results stored in the cell metadata.
    """

    def replace_variables(self, source: str, variables: Dict[str, str]) -> str:
        """Replace {{variable-name}} with stored value."""
        try:
            replaced = re.sub(
                "{{(.*?)}}", lambda m: variables.get(m.group(1), ""), source
            )
        except TypeError:
            replaced = source
        return replaced

    def preprocess_cell(
        self, cell: "NotebookNode", resources: Dict[str, bool], index: int
    ) -> Tuple["NotebookNode", Dict[str, bool]]:
        """Preprocess cell.

        Parameters
        ----------
        cell : NotebookNode cell
            Notebook cell being processed
        resources : dictionary
            Additional resources used in the conversion process. Allows
            preprocessors to pass variables into the Jinja engine.
        cell_index : int
            Index of the cell being processed (see base.py)

        """
        if cell.cell_type == "markdown":
            variables = cell["metadata"].get("variables", {})
            if len(variables) > 0:
                cell.source = self.replace_variables(cell.source, variables)
                if resources.get("delete_pymarkdown", False):
                    del cell.metadata["variables"]
        return cell, resources



"""Test the extract_attachments preprocessor."""
from nbformat.v4 import new_markdown_cell
from binascii import a2b_base64

from thermohw import ExtractAttachmentsPreprocessor


preproc = ExtractAttachmentsPreprocessor()


def test_pathological_image_name() -> None:
    """Test that image name with escaped characters is converted properly."""
    fname = (
        "%28%29%40%23%24%25%5E%26%2A%21%3C%3E%3F%2C%7B%7D%5B%5D%5C%7C%7E%60%2B%3D%20"
    )
    repl = "-" * (len(fname) // 3)
    data = (
        "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB"
        "0C8AAAAASUVORK5CYII="
    )
    in_cell = new_markdown_cell(source=(f"![{fname}.png](attachment:{fname}.png)"))
    in_cell["attachments"] = {f"{fname}.png": {"image/png": data}}
    cell, resources = preproc.preprocess_cell(in_cell, {"outputs": {}}, 0)
    assert cell.source == f"![{fname}.png](_0_{repl}.png)"
    assert resources["outputs"][f"_0_{repl}.png"] == a2b_base64(data)



"""Convert divs with alert classes to equivalent LaTeX environments.

This module uses the JSON output format from Pandoc to look for ``Div``
elements (which are the Pandoc AST representation of ``div`` elements)
and convert them to a LaTeX environment with the appropriate name.
"""

from enum import Enum, auto
from typing import Any, Optional

from pandocfilters import applyJSONFilters, RawBlock, RawInline


# Allowed alert types are all from the Bootstrap alert types
# https://getbootstrap.com/docs/4.1/components/alerts/
class ALLOWED_ALERT_TYPES(Enum):
    success = auto()
    primary = auto()
    secondary = auto()
    warning = auto()
    danger = auto()
    info = auto()


def div_filter(key: str, value: list, format: str, meta: Any) -> Optional[list]:
    """Filter the JSON ``value`` for alert divs.

    Arguments
    ---------
    key
        Key of the structure
    value
        Values in the structure
    format
        Output format of the processing
    meta
        Meta information
    """
    if key != "Div" or format != "latex":
        return None

    [[_, classes, _], contents] = value
    try:
        alert_type = [name.split("-")[1] for name in classes if "-" in name][0]
    except IndexError:
        return None

    if alert_type not in ALLOWED_ALERT_TYPES.__members__:
        return None

    filtered = [RawBlock("latex", rf"\begin{{{alert_type}box}}")]
    filtered.extend(contents)
    filtered.append(RawBlock("latex", rf"\end{{{alert_type}box}}"))
    return filtered


def convert_div(text: str, format: Optional[str] = None) -> "applyJSONFilters":
    """Apply the `div_filter` action to the text."""
    return applyJSONFilters([div_filter], text, format=format)


def raw_html_filter(key: str, value: list, format: str, meta: Any) -> Optional[list]:
    """Filter the JSON ``value`` for raw html to convert to LaTeX.

    Arguments
    ---------
    key
        Key of the structure
    value
        Values in the structure
    format
        Output format of the processing
    meta
        Meta information
    """
    if key == "RawInline" and format == "latex" and value[0] == "html":
        if value[1] == "<sup>":
            filtered = [RawInline("latex", r"\textsuperscript{")]
        elif value[1] == "</sup>":
            filtered = [RawInline("latex", "}")]
        elif value[1] == "<sub>":
            filtered = [RawInline("latex", r"\textsubscript{")]
        elif value[1] == "</sub>":
            filtered = [RawInline("latex", "}")]
        else:
            return None
        return filtered

    return None


def convert_raw_html(text: str, format: Optional[str] = None) -> "applyJSONFilters":
    """Apply the `raw_html_filter` action to the text."""
    return applyJSONFilters([raw_html_filter], text, format=format)



from setuptools import setup
from pathlib import Path
from typing import Dict

HERE = Path(__file__).parent

version: Dict[str, str] = {}
with HERE.joinpath("src", "thermohw", "_version.py").open(mode="r") as version_file:
    exec(version_file.read(), version)

setup(version=version["__version__"], package_dir={"": "src"})



"""
Utility functions for the converters.
"""

# Standard Library
from typing import List
from io import BytesIO

# Third-Party
from pdfrw import PdfReader, PdfWriter


def combine_pdf_as_bytes(pdfs: List[BytesIO]) -> bytes:
    """Combine PDFs and return a byte-string with the result.

    Arguments
    ---------
    pdfs
        A list of BytesIO representations of PDFs

    """
    writer = PdfWriter()
    for pdf in pdfs:
        writer.addpages(PdfReader(pdf).pages)
    bio = BytesIO()
    writer.write(bio)
    bio.seek(0)
    output = bio.read()
    bio.close()
    return output



"""Preprocessors for homework assignments.

This module contains the preprocessors that are used to convert
Notebooks into homework assignments.

Classes
-------
HomeworkPreprocessor:
    Preprocess a homework problem for notes and other things that don't
    belong in an assignment.

SolnRemoverPreprocessor:
    Preprocess the Notebook to remove the solution section and replace
    it with headings for solution parts.

"""

# Standard Library
from typing import TYPE_CHECKING, Tuple, List, Dict
import warnings

# Third-Party
from nbconvert.preprocessors import Preprocessor
from nbformat.v4 import new_code_cell, new_markdown_cell

if TYPE_CHECKING:
    from nbformat import NotebookNode  # noqa: F401 # typing only


by_hand_source = (
    "**Attach an image of your solution for this problem in this cell. "
    "Attach multiple images, one in each cell, if necessary. Please make "
    "sure the text is clear and legible.**"
)
by_hand_cell = new_markdown_cell(source=by_hand_source)

md_expl_source = (
    "**Write your engineering model, equations, and/or explanation of your process "
    "here.**"
)
md_expl_cell = new_markdown_cell(source=md_expl_source)

code_ans_source = (
    "# Write your code here to solve the problem\n"
    "# Make sure to write your final answer in the cell below."
)
code_ans_cell = new_code_cell(source=code_ans_source)

md_ans_source = """\
<div class="alert alert-success">

**Answer:**

</div>
"""
md_ans_cell = new_markdown_cell(source=md_ans_source)
md_ans_cell.metadata.deletable = False

sketch_source = "**Attach an image of your sketch for this problem in this cell.**"
sketch_cell = new_markdown_cell(source=sketch_source)


class RawRemover(Preprocessor):  # type: ignore
    """Remove any raw cells from the Notebook."""

    def preprocess(
        self, nb: "NotebookNode", resources: Dict[str, Dict[str, bool]]
    ) -> Tuple["NotebookNode", Dict[str, Dict[str, bool]]]:
        """Remove any raw cells from the Notebook.

        By default, exclude raw cells from the output. Change this by including
        global_content_filter->include_raw = True in the resources dictionary.
        This preprocessor is necessary because the NotebookExporter doesn't
        include the exclude_raw config."""
        if not resources.get("global_content_filter", {}).get("include_raw", False):
            keep_cells = []
            for cell in nb.cells:
                if cell.cell_type != "raw":
                    keep_cells.append(cell)

            nb.cells = keep_cells

        return nb, resources


class SolutionRemover(Preprocessor):  # type: ignore
    """Preprocess a homework problem to remove the solution.

    This preprocessor produces output suitable for distribution to students as
    an assignment by removing the solution from the Notebook.

    The solution cells are found by tags attached to the cells. A cell tagged
    ``solution`` will start the Solution section. Following that, any cell
    tagged ``part`` will insert a solution prompt into the assignment. If the
    additional tag ``sketch`` is added, then a prompt for a sketch is inserted
    instead. Prior to the start of the Solution, all cells are kept.

    To enable the legacy processing behavior, described below, the
    resources->legacy key must be set to True. If no legacy key is present in
    resources, the default is assumed to be True for now. This will change in
    the future.

    The preprocessor looks for a cell with the header '## solution'
    (case-insensitive) in it, which delimits the beginning of the solution
    section. Then, for every cell after that, it checks for a level 3 header,
    which delimits the start of a section of the solution. All cells between the
    level 3 headers are replaced with cells that ask the students to write their
    code and explanation.

    The processing is only done if the resources->remove_solution key is True.
    """

    def preprocess(
        self, nb: "NotebookNode", resources: Dict[str, bool]
    ) -> Tuple["NotebookNode", Dict[str, bool]]:
        """Preprocess the entire notebook."""
        if "remove_solution" not in resources:
            raise KeyError("The resources dictionary must have a remove_solution key.")
        if not resources["remove_solution"]:
            return nb, resources

        if resources.get("legacy", True):
            warnings.warn(
                "The legacy behavior for finding solution cells will be removed in "
                "the future.",
                FutureWarning,
            )
            return self.legacy_parser(nb, resources)

        keep_cells: List["NotebookNode"] = []
        solution_started = False
        for cell in nb.cells:
            if "tags" in cell.metadata:
                tags = cell.metadata["tags"][:]
                del cell.metadata["tags"]
            else:
                tags = []
            if "solution" in tags:
                solution_started = True
                keep_cells.append(cell)
            elif "part" in tags:
                keep_cells.append(cell)
                if "sketch" in tags:
                    keep_cells.append(sketch_cell)
                elif resources["by_hand"]:
                    keep_cells.append(by_hand_cell)
                else:
                    keep_cells.append(md_expl_cell)
                    keep_cells.append(code_ans_cell)
                    keep_cells.append(md_ans_cell)
            else:
                if tags:
                    warnings.warn(f"Unknown tag value: {tags}", UserWarning)
                if not solution_started:
                    keep_cells.append(cell)

        nb.cells = keep_cells
        return nb, resources

    def legacy_parser(
        self, nb: "NotebookNode", resources: Dict[str, bool]
    ) -> Tuple["NotebookNode", Dict[str, bool]]:
        keep_cells_idx: List[int] = []
        for index, cell in enumerate(nb.cells):
            if "## solution" in cell.source.lower():
                keep_cells_idx.append(index)
            # The space at the end of the test string here is important
            elif len(keep_cells_idx) > 0 and cell.source.startswith("### "):
                keep_cells_idx.append(index)

        if len(keep_cells_idx) == 0:
            raise ValueError(
                "The Notebook must have a cell with the text '## solution' "
                "(case insensitive) in it."
            )

        keep_cells = nb.cells[: keep_cells_idx[0] + 1]
        if len(keep_cells_idx) == 1:
            if resources["by_hand"]:
                keep_cells.append(by_hand_cell)
            else:
                if "sketch" in nb.cells[keep_cells_idx[0]].source.lower():
                    keep_cells.append(sketch_cell)
                else:
                    keep_cells.append(md_expl_cell)
                    keep_cells.append(code_ans_cell)
                    keep_cells.append(md_ans_cell)
        else:
            for i in keep_cells_idx[1:]:
                keep_cells.append(nb.cells[i])
                if resources["by_hand"]:
                    keep_cells.append(by_hand_cell)
                else:
                    if "sketch" in nb.cells[i].source.lower():
                        keep_cells.append(sketch_cell)
                    else:
                        keep_cells.append(md_expl_cell)
                        keep_cells.append(code_ans_cell)
                        keep_cells.append(md_ans_cell)

        nb.cells = keep_cells
        return nb, resources
